"""Util module.

The util module holds additional files which are used in this framework/
library but have no main functionality components. These include plotting
functions, logging functionality and other helper functions.
"""
