"""This is the ReLeSO package. This is the best!

Please view the Documentation under :doc:`/index` for the information
about this library and framework.
"""
from releso.__version__ import __version__ as __version__
